# -*- coding: utf-8 -*-
"""Miscellaneous utilities.
"""

__all__ = ('add_ro_dict_property',
           'ensure_kwarg_in',
           'ensure_kwarg_not_in',
           'check_kwarg',
           'check_kwargs')


def _dict_property_doc(locs, kw):
    try:
        tdoc = kw['doc']
    except KeyError:
        doc = None
    else:
        doc = tdoc % dict(locs, **kw)
    return doc


def _dict_property_fget(dictattr, key, default):
    if callable(default):
        return lambda obj, k=key: getattr(obj, dictattr).get(k, default(obj))
    else:
        return lambda obj, k=key: getattr(obj, dictattr).get(k, default)


def add_ro_dict_property(cls, dictattr, attr, default=None, **kw):
    """Add to class **cls** a read-only property returning a predefined entry
       from a dict.

       :param type cls:
            the target class to be modified,
       :param str dictattr:
            name of the attribute of **cls** being the source dict,
       :param str,tuple attr:
            name of the generated attribute, or a tuple ``(attr, key)``,
            where ``attr`` is the generated attribute name and ``key`` is
            the corresponding key in the dictionary,
       :param default:
            default value returned by the property, if the dictionary
            has no requested key, if **default** is callable, it shall
            be a function ``default(obj)``, where an instance of **cls**
            will be passed as ``obj``,
       :keyword str doc:
            optional documentation string, the string may use substitutions,
            such as ``"%(cls)s"``, ``"%(dictattr)s"``, ``"%(attr)s"``,
            ``"%(key)s"``, ``"%(default)s"``. Extra variables for necessary
            substitutions may be provided via keyword arguments ``**kw``.
    """
    if isinstance(attr, str):
        key = attr
    else:
        (attr, key) = attr  # assumed it's a tuple or alike

    fget = _dict_property_fget(dictattr, key, default)
    locs = {k: v for k, v in locals().items() if k not in ('kw', 'fget')}
    doc = _dict_property_doc(locs, kw)
    setattr(cls, attr, property(fget, doc=doc))


def ensure_kwarg_in(caller, key, allowed):
    """Checks a single **key** from keyword arguments against allowed keys.

       If **key** is not in **allowed** keys, throws :exc:`TypeError`.

       :param str caller: name of the caller,
       :param str key: the key to be examined,
       :param allowed: sequence of allowed keys.

       :return: True
    """
    if allowed is not None and key not in allowed:
        msg = "%s got an unexpected keyword argument %r" % \
                (str(caller), key)
        raise TypeError(msg)
    return True


def ensure_kwarg_not_in(caller, key, forbidden):
    """Checks a single **key** from keyword arguments against forbidden keys.

       If **key** is in **forbidden** keys, throws :exc:`TypeError`.

       :param str caller: name of the caller,
       :param str key: the key to be examined,
       :param forbidden: sequence of forbidden keys.

       :return: True
    """
    if forbidden is not None and key in forbidden:
        msg = "%s got a forbidden keyword argument %r" % \
                (str(caller), key)
        raise TypeError(msg)
    return True


def check_kwarg(caller, key, allowed=None, forbidden=None):
    """Checks a single **key** from keyword arguments against allowed and
       forbidden keys.

       If **key** is not in **allowed** keys or it is in **forbidden** keys,
       throws :exc:`TypeError`.

       :param str caller: name of the caller,
       :param str key: the key to be examined,
       :param allowed: sequence of allowed keys.
       :param forbidden: sequence of forbidden keys.

       :return: True
    """
    ensure_kwarg_in(caller, key, allowed)
    ensure_kwarg_not_in(caller, key, forbidden)
    return True


def check_kwargs(caller, kw, allowed=None, forbidden=None):
    """Checks all keys from keyword arguments **kw** against allowed and
       forbidden keys.

       If **key** is not in **allowed** keys or it is in **forbidden** keys,
       throws :exc:`TypeError`.

       :param str caller: name of the caller,
       :param str key: the key to be examined,
       :param allowed: sequence of allowed keys.
       :param forbidden: sequence of forbidden keys.

       :return: True
    """
    for key in kw:
        check_kwarg(caller, key, allowed, forbidden)
    return True


# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set ft=python et ts=4 sw=4:



# -*- coding: utf-8 -*-
"""Provides the :class:`.Replacements` class.
"""


__all__ = ('Replacements',
           'ReplacingCaller',
           'ReplacingBuilder',
           'ReplacingAction')


_scons_env_setters = ('SetDefault',
                      'Replace',
                      'Append',
                      'AppendUnique',
                      'Prepend',
                      'PrependUnique')


def _is_scons_env(obj):
    return all(hasattr(obj, m) for m in _scons_env_setters)


def _method_and_name(obj, method):
    if isinstance(method, str):
        name = method
        method = getattr(obj, method)
    else:
        name = method.__name__
    return (method, name)


class Replacements(dict):
    """Enables one to temporarily replace variables in a dict-like object
    (SCons ``Environment``).

    The :class:`.Replacements` are designed to be used with
    :class:`.ReplacingBuilder` or :class:`.ReplacingAction`.
    """
    def mapped_variables(self, only=None):
        if only is not None:
            return {v: '$' + k for k, v in self.items() if k in only}
        else:
            return {v: '$' + k for k, v in self.items()}

    def inject(self, dest, setter='__setitem__', only_present=False):
        """Inject replacement variables into the **dest**.

        This method is supposed to be used to initialize replacement variables
        in the dictionary **dest**. For each mapping ``{'FOO': 'BAR'}`` in
        :class:`.Replacements` ``dest['BAR'] = '$FOO'`` assignment will be
        performed using the provided **setter**. If **only_present** is
        ``True``, the assignment will be performed only if the variable
        ``dest['FOO']`` is already defined.

        :param dict dest:
            destination dictionary to be modified,
        :param str,callable setter:
            a method (or method name) of **dest** that sets items in **dest**,
            should be a function with prototype ``dest.setter(name, value)``,
            except for ``SetDefault``, ``Replace``, ``Append``, ``Prepend``,
            ``AppendUnique``, ``PrependUnique`` methods of SCons environment
            where the prototype is ``dest.setter(**kw)``,
        :param bool only_present:
            if ``True``, only the replacements for variables already present in
            **dest** will be set.
        """
        variables = self.mapped_variables(dest if only_present else None)
        self._do_inject(dest, setter, variables)

    def apply(self, subj, include_unmapped=False):
        """Apply replacements to **subj**.

        Returns a dictionary of variables from **subj** set to their
        replacement values. By default, only variables having their
        replacements are extracted. If **include_unmapped** is ``True``,
        remaining variables from **subj** will also be added to the
        returned dictionary.

        :param dict subj:
            a dictionary with variables, some of which may be replacement
            variables,
        :param bool include_unmapped:
            also include the variables from **subj** which are not replacement
            variables.
        :return dict: replaced variables from *subj**.
        """
        def selfref(k, v): return subj.get(v, '$' + k) == '$' + k
        variables = {k: subj[v] for k, v in self.items() if not selfref(k, v)}
        if include_unmapped:
            mapped = set(self.values()) | set(variables)
            variables.update({k: subj[k] for k in subj if k not in mapped})
        return variables

    def _do_inject(self, dest, setter, variables):
        setter, setter_name = _method_and_name(dest, setter)
        if _is_scons_env(dest) and setter_name in _scons_env_setters:
            setter(**variables)
        else:
            for k, v in variables.items():
                setter(k, v)


class ReplacingCaller(object):
    """Base class for :class:`.ReplacingBuilder`, :class:`.ReplacingAction`
    and other similar wrappers.

    """
    def __init__(self, wrapped, replacements=dict(), **kw):
        """
        :param wrapped: callable object, initializes :attr:`.wrapped`,
        :param dict replacements: initializes :attr:`.replacements`,
        :param kw: keyword arguments used to initialize :attr:`.replacements`.
        """
        #: Wrapped callable that will be invoked with replaced variables.
        self.wrapped = wrapped
        #: An instance of :class:`.Replacements` used to replace variables.
        self.replacements = Replacements(replacements, **kw)

    def _wrapper_attributes(self):
        return ('wrapped',
                'replacements',
                'apply_replacements',
                'inject_replacements',
                '_call',
                'sort_call_args')

    def __getattr__(self, name):
        """Provides read acces to attributes of :attr:`.wrapped`."""
        return getattr(self.wrapped, name)

    def __setattr__(self, name, value):
        """Provides write access to attribtes of :attr:`.wrapped`."""
        if name in ('_wrapper_attributes',):
            raise AttributeError("can't set attribute")
        elif name in self._wrapper_attributes():
            super(ReplacingCaller, self).__setattr__(name, value)
        else:
            setattr(self.wrapped, name, value)

    def __call__(self, env, *args, **kw):
        """Invokes :attr:`.wrapped` with **args** and replaced variables in
        **env** and **kw**.

        :param env: SCons environment, an ``env.Override(...)`` will be passed
                    to :attr:`.wrapped` instead of ``env``,
        :param args: arguments to be passed to :attr:`.wrapped`,
        :param kw: keyword args (may be used to override variables in env).
        """
        return self._call(env, *args, **kw)

    def _call(self, env, *args, **kw):
        (env, kw) = self.apply_replacements(env, **kw)
        return self.wrapped(*self.sort_call_args(env, *args), **kw)

    def apply_replacements(self, env, **kw):
        """Applies replacements to env and kw."""
        ovr = self.replacements.apply(env)
        kw = self.replacements.apply(kw, True)
        return (env.Override(ovr), kw)

    def inject_replacements(self, env, setter='__setitem__',
                            only_present=False):
        """Same as :meth:`replacements.inject(self,env,setter,only_present)
        <.Replacements.inject>`."""
        self.replacements.inject(self, env, setter, only_present)

    def sort_call_args(self, *args):
        """May be be overwritten in a subclass to reorganize positional
        arguments passed to :attr:`.wrapped`."""
        return args


class ReplacingBuilder(ReplacingCaller):
    """SCons builder wrapper, calls the wrapped builder with replaced
    construction variables.

    :Example: Typical usage

    The following builder named ``MyObject`` is similar to SCons ``Object``
    builder, but it uses ``MY_CFLAGS`` instead of ``CFLAGS``.

    .. code-block:: python

        from sconstool.util import ReplacingBuilder
        from SCons.Environment import Environment

        env = Environment(tools=['default'], MY_CFLAGS=['-Wall', '-Wextra'])
        bld = ReplacingBuilder(env['BUILDERS']['Object'], CFLAGS='MY_CFLAGS')
        env['BUILDERS']['MyObject'] = bld
        bld.inject_replacements(env, 'SetDefault')

        env.MyObject('test1.c') # gcc -c -o test1.o -Wall -Wextra test1.c
        env.Object('test2.c')   # gcc -c -o test2.o test2.c
    """
    def __call__(self, env, target, source, *args, **kw):
        """Calls the wrapped builder with replaced construction variables."""
        return ReplacingCaller._call(self, env, target, source, *args, **kw)


class ReplacingAction(ReplacingCaller):
    """SCons action wrapper, replaces construction variables and calls the
    wrapped action."""

    def sort_call_args(self, env, target, source, *args):
        return (target, source, env) + args

    def __call__(self, target, source, env, *args, **kw):
        """Calls the wrapped action with replaced construction variables."""
        return ReplacingCaller._call(self, env, target, source, *args, **kw)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set ft=python et ts=4 sw=4:



# -*- coding: utf-8 -*-
"""Utility functions to automate (sub)module imports.
"""

import importlib
import inspect
import sys

__all__ = ('import_all_from', 'import_from')


def import_all_from(target, modules, module_package=None, **kw):
    """Imports symbols from multiple modules.

       For each **module** in **modules**, this is an equivalent of

       .. code-block:: python

            from module import *
            __all__ += module.__all__

       :param module|str target:
            target module or package, to which the **modules** have to be
            imported,
       :param list|str modules:
            modules to be imported,
       :param module_package:
            package of the module being imported, if ``None`` (default),
            the **target** is used as module_package.
    """
    if isinstance(modules, str) or inspect.ismodule(modules):
        modules = (modules,)
    target_module, target = _get_module_and_name(target)
    for module in modules:
        if isinstance(module, str):
            module = importlib.import_module(module, module_package or target)
        import_from(target_module, module, all_symbols(module), **kw)


def import_from(target, source, symbols, **kw):
    for symbol in symbols:
        setattr(target, symbol, getattr(source, symbol))
    if kw.get('__all__', True):
        if not hasattr(target, '__all__'):
            target.__all__ = ()
        target.__all__ += target.__all__.__class__(symbols)


def all_symbols(module):
    if hasattr(module, '__all__'):
        return module.__all__
    else:
        return dir(module)


def _get_module_and_name(mod):
    """Return a module object and its name for module given as either a name
       or a module object."""
    if isinstance(mod, str):
        return (sys.modules[mod], mod)
    else:  # module, or any other object featuring __name__ (like class, e.g)
        return (mod, mod.__name__)


# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set ft=python et ts=4 sw=4:



# -*- coding: utf-8 -*-
"""Provides the :class:`.ToolFinder` class.
"""

from . import misc_
import os


__all__ = ('ToolFinder',)


class ToolFinder(object):
    """Callable object which searches for executables.

    A single ToolFinder instance searches for a single file (program), for
    example a compiler executable or script interpreter. The constructor
    accepts several options, for each option there is corresponding
    @property (read-only) with the same name.

    :Example: Typical use in a tool module

    .. code-block:: python

        from sconstool.util import ToolFinder
        foo = ToolFinder('foo')

        def generate(env):
            env.SetDefault(FOO=foo(env))
            # ...

        def exists(env):
            return env.get('FOO', foo(env))
    """
    __slots__ = ('_tool', '_kw')

    _ctor_kwargs = ('name',
                    'path',
                    'pathext',
                    'reject',
                    'priority_path',
                    'fallback_path',
                    'strip_path',
                    'strip_priority_path',
                    'strip_fallback_path')

    def __init__(self, tool, **kw):
        """
        :param str tool:
            symbolic name of the tool,
        :keyword str,list name:
            base name of the file (program name) being searched for,
            may be a list of alternative program names,
        :keyword str,list path:
            search path to be used instead of the standard SCons PATH,
        :keyword str,list pathext:
            a list of file extensions to be considered as executable,
        :keyword list reject:
            a list of paths to be rejected,
        :keyword str,list priority_path:
            extra search path to be searched prior to :attr:`.path`,
        :keyword str,list fallback_path:
            extra search path to be searched after :attr:`.path`,
        :keyword bool strip_path:
            if ``True`` (default), the leading path, if it's in :attr:`path`
            list, will be stripped from the returned file path,
        :keyword bool strip_priority_path:
            if ``True``, the leading path, if it's in **priority_path**
            list, will be stripped from the returned file path;
        :keyword bool strip_fallback_path:
            if ``True``, the leading path, if it's in **fallback_path** list,
            will be stripped from the returned file path.
        """
        self._tool = str(tool)
        misc_.check_kwargs('ToolFinder()', kw, self._ctor_kwargs)
        self._kw = kw

    @property
    def tool(self):
        """Tool name, that was passed in to the c-tor as an argument.

           :rtype: str
        """
        return self._tool

    def __call__(self, env):
        """Performs the actual search.

           :param env:
                a SCons environment; provides construction variables and the
                ``env.WhereIs()`` method to the :class:`.ToolFinder`.
           :return:
                depending on options chosen at object creation, a name or a
                path to the executable file found. If the program can't be
                found, ``None`` is returned.
           :rtype: str
        """
        return self._search(env)

    def _whereis(self, env, prog, where):
        path = getattr(self, where)
        if path and not isinstance(path, str):
            # this trick enables variable substitution in list entries
            path = os.path.pathsep.join(path)
        return env.WhereIs(prog, path, self.pathext, self.reject)

    def _adjust_result(self, env, result, where):
        prog = env.subst(result[0])
        strip = getattr(self, 'strip_%s' % where)
        if os.path.isabs(prog) or strip:
            return prog
        return result[1]

    def _search_in(self, env, where):
        progs = self.name
        if isinstance(progs, str):
            progs = [progs]
        for prog in progs:
            found = self._whereis(env, prog, where)
            if found:
                return self._adjust_result(env, (prog, found), where)
        return None

    def _search(self, env):
        for where in ('priority_path', 'path', 'fallback_path'):
            found = self._search_in(env, where)
            if found:
                return found
        return None

    @classmethod
    def _add_getter(cls, attr, default=None, **kw):
        if isinstance(default, property):
            default = default.fget
            kw['defaultattr'] = default.__name__

            doc = """\
            The value of **%(attr)s** keyword argument passed in to the
            constructor at object creation, or ``self.%(defaultattr)s`` if the
            argument was omitted.

            :rtype: %(rtype)s
            """
        else:
            doc = """\
            The value of **%(attr)s** keyword argument passed in to the
            constructor at object creation, or ``%(default)r`` if the
            argument was omitted.

            :rtype: %(rtype)s
            """
        kw = dict({'doc': doc}, **kw)
        misc_.add_ro_dict_property(cls, '_kw', attr, default, **kw)


TF = ToolFinder
TF._add_getter('name', TF.tool, rtype='str')
TF._add_getter('path', rtype='str,list')
TF._add_getter('priority_path', [], rtype='str,list')
TF._add_getter('fallback_path', [], rtype='str,list')
TF._add_getter('pathext', rtype='str,list')
TF._add_getter('reject', [], rtype='list')
TF._add_getter('strip_path', True, rtype='bool')
TF._add_getter('strip_priority_path', False, rtype='bool')
TF._add_getter('strip_fallback_path', False, rtype='bool')
del TF


# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set ft=python et ts=4 sw=4:



# -*- coding: utf-8 -*-
"""Provides the :class:`.Selector` class.
"""


__all__ = ('Selector',)


class Selector(dict):
    """A replacement for ``SCons.Util.Selector`` class.

    :class:`.Selector` returns a dictionary value for the longest matched
    source suffix found in the dictionary.

    To have an example, for the following selector

    .. code-block:: python

            sel = Selector({'.h': 'H', '.t.h': 'TH'})

    all the assertions below hold

    .. code-block:: python

            assert sel(env, ['foo.h']) == 'H'
            assert sel(env, ['foo.t.h']) == 'TH'

    The original ``SCons.Util.Selector`` would only select ``'H'``, no matter
    what you do.
    """
    def __call__(self, env, source, ext=None):
        select_from = _get_selector_func(source, ext)
        items = []
        for x_dict in _separate_literals(self.items(), env):
            try:
                item = select_from(x_dict)
            except KeyError:
                pass
            else:
                items.append(item)
        try:
            return _choose_better(items)
        except IndexError:
            return self.get(None)


def _get_selector_func(source, ext):
    if ext is not None:
        return lambda x_dict, e=ext: (e, x_dict[e])
    try:
        src = str(source[0])
    except IndexError:
        return lambda x_dict: ('', x_dict[''])
    else:
        return lambda x_dict, s=src: _select_best(x_dict.items(), s)


def _separate_literals(items, env):
    # Split-up items into two dictionaries. First one with items whose keys
    # were given literally, and the second with items whose keys had
    # substitutions.
    (l_dict, s_dict) = ({}, {})
    for item in items:
        _separate_handle_item(env, (l_dict, s_dict), item)
    return (l_dict, {s_k: x[1] for (s_k, x) in s_dict.items()})


def _separate_handle_item(env, dicts, item):
    (l_dict, s_dict) = dicts
    (k, v) = item
    if k is None:
        return
    s_k = env.subst(k)
    if k == s_k:
        l_dict[k] = v  # it's literal
    else:
        if s_k in s_dict:
            # We only raise an error when variables point
            # to the same suffix. If one suffix is literal
            # and a variable suffix contains this literal,
            # the literal wins and we don't raise an error.
            raise KeyError(s_dict[s_k][0], k, s_k)
        s_dict[s_k] = item


def _select_best(items, src):
    ml = [(suf, v) for (suf, v) in items if src.endswith(suf)]
    if not ml:
        raise KeyError('suffix not found for %s' % repr(src))
    ml = list(sorted(ml, key=lambda x: len(x[0]), reverse=True))
    return ml[0]


def _choose_better(items):
    (k, v) = items[0]
    if len(items) == 2 and len(items[1][0]) > len(k):
        (k, v) = items[1]
    return v


# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set ft=python et ts=4 sw=4:



# -*- coding: utf-8 -*-
"""Top-level module."""

from .about import __version__

from .imports_ import *

__all__ = imports_.__all__

import_all_from(__package__, [
    '.misc_',
    '.finder_',
    '.emitter_',
    '.selector_',
    '.replacements_'
])

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set ft=python et ts=4 sw=4:



# -*- coding: utf-8 -*-
"""Provides the :class:`.ConditionalEmitter` class.
"""


__all__ = ('ConditionalEmitter',)


class ConditionalEmitter(object):
    """A callable object, which calls user-provided **emitter** when a
    predefined condition is meet."""

    __slots__ = ('_predicate', '_emitter_if', '_emitter_else')

    def __init__(self, predicate, emitter_if=None, emitter_else=None):
        """
        :param predicate:
            a callable object (function) of type ``pred(target, source, env)``;
            the function should return a boolean value; if ``pred()`` returns
            true, the :class:`.ConditionalEmitter`` will dispatch control to
            **emitter_if**, otherwise it will call **emitter_else**,
        :param emitter_if:
            an emitter function of the form ``emitter(target, source, env)``
            which gets called when ``pred()`` returns ``True``,
        :param emitter_else:
            an emitter function of the form ``emitter(target, source, env)``
            which gets called when ``pred()`` returns ``False``,
        """
        if not callable(predicate):
            raise TypeError("predicate must be callable")
        self._predicate = predicate
        self._emitter_if = emitter_if
        self._emitter_else = emitter_else

    @property
    def predicate(self):
        """The value of **predicate** parameter passed in to the constructor
        at object creation"""
        return self._predicate

    @property
    def emitter_if(self):
        """The value of **emitter_if** parameter passed in to the constructor
        at object creation, or :attr:`.default_emitter` if **emitter_if** was
        omitted."""
        return self._emitter_if or self.default_emitter

    @property
    def emitter_else(self):
        """The value of **emitter_else** parameter passed in to the constructor
        at object creation, or :attr:`.default_emitter` if **emitter_else** was
        omitted."""
        return self._emitter_else or self.default_emitter

    def default_emitter(self, target, source, env):
        """Default emitter, just returns the tuple ``(target, source)``."""
        return (target, source)

    def __call__(self, target, source, env):
        cond = self.predicate(target, source, env)
        emitter = self.emitter_if if cond else self.emitter_else
        return emitter(target, source, env)


# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set ft=python et ts=4 sw=4:



# -*- coding: utf-8 -*-
"""scons-tool-util
"""

from setuptools import setup
from os import path
from sys import version_info


here = path.abspath(path.dirname(__file__))

if version_info < (3, 0):
    from io import open as uopen
else:
    uopen = open

readme_rst = path.join(here, 'README.rst')
with uopen(readme_rst, encoding='utf-8') as f:
    readme = f.read()

about = {}
about_py = path.join(here, 'lib', 'sconstool', 'util', 'about.py')
with open(about_py) as f:
    exec(f.read(), about)

setup(
        name='scons-tool-util',
        version=about['__version__'],
        package_dir={'': 'lib'},
        packages=['sconstool.util'],
        namespace_packages=['sconstool'],
        description='A library of utility functions and objects for ' +
                    'scons-tool-* packages',
        long_description=readme,
        long_description_content_type='text/x-rst',
        url='https://github.com/ptomulik/scons-tool-util',
        author='Paweł Tomulik',
        author_email='ptomulik@meil.pw.edu.pl'
)

# vim: set expandtab tabstop=4 shiftwidth=4:
